@@ -18,7 +18,6 @@ import time
 import typing
 import uuid
 import warnings
-from collections import OrderedDict
 from itertools import groupby
 from pprint import pprint
 from textwrap import indent
@@ -29,24 +28,20 @@ import bluesky.plans as bp
 import bluesky.preprocessors as bpp
 import yaml
 from bluesky import RunEngine
-from bluesky.callbacks.best_effort import BestEffortCallback
 from bluesky.callbacks.broker import verify_files_saved
+from bluesky.preprocessors import msg_mutator
 from bluesky.preprocessors import pchain
 from bluesky.suspenders import SuspendFloor
-from bluesky.utils import Msg, normalize_subs_input, single_gen
+from bluesky.utils import normalize_subs_input, single_gen, Msg
 from ophyd import Device
+from xpdconf.conf import XPD_SHUTTER_CONF
 
-from xpdacq.beamtime import (Beamtime, ScanPlan, close_shutter_stub,
-                             open_shutter_stub)
+from xpdacq.beamtime import Beamtime, ScanPlan
+from xpdacq.beamtime import close_shutter_stub, open_shutter_stub
 from xpdacq.glbl import glbl
-from xpdacq.preprocessors import (CalibPreprocessor, DarkPreprocessor,
-                                  MaskPreprocessor, ShutterPreprocessor)
-from xpdacq.tools import xpdAcqError, xpdAcqException
-from xpdacq.xpdacq_conf import XPDACQ_MD_VERSION, xpd_configuration
+from xpdacq.tools import xpdAcqException
+from xpdacq.xpdacq_conf import xpd_configuration, XPDACQ_MD_VERSION
 
-Plan = typing.Generator[Msg, typing.Any, typing.Any]
-MaskFiles = typing.List[typing.Tuple[Device, typing.List[str]]]
-PoniFile = typing.List[typing.Tuple[Device, str]]
 XPD_shutter = xpd_configuration.get("shutter")
 PAUSE_MSG = """
 Your RunEngine (xrun) is entering a paused state.
@@ -75,9 +70,8 @@ def periodic_dark(plan):
 
         if (not need_dark) and (not qualified_dark_uid):
             need_dark = True
-        if need_dark and (not qualified_dark_uid) and msg.command == "open_run" and (
-            "dark_frame" not in msg.kwargs
-        ):
+        if need_dark and (not qualified_dark_uid) and msg.command == "open_run" and \
+                ("dark_frame" not in msg.kwargs):
             # We are about to start a new 'run' (e.g., a count or a scan).
             # Insert a dark frame run first.
             need_dark = False
@@ -148,16 +142,6 @@ class CustomizedRunEngine(RunEngine):
         super().__init__(*args, **kwargs)
         self._beamtime = beamtime
         self.pause_msg = PAUSE_MSG
-        self.dark_preprocessors: typing.List[DarkPreprocessor] = []
-        self.calib_preprocessors: typing.List[CalibPreprocessor] = []
-        self.shutter_preprocessors: typing.List[ShutterPreprocessor] = []
-        bec = BestEffortCallback()
-        bec.noplot_streams.extend(["calib", "dark"])
-        bec.disable_baseline()
-        bec.disable_plots()
-        bec.enable_heading()
-        bec.enable_table()
-        self.subscribe(bec)
 
     @property
     def beamtime(self):
@@ -181,80 +165,14 @@ class CustomizedRunEngine(RunEngine):
         exp_hash_uid = str(uuid.uuid4())
         glbl["exp_hash_uid"] = exp_hash_uid
 
-    def _make_cpps(self, poni_file: PoniFile):
-        cpps = []
-        for det, poni in poni_file:
-            cpp = CalibPreprocessor(det)
-            cpp.load_calib_result({}, poni)
-            cpps.append(cpp)
-        return cpps
-
-    def _make_mpps(self, mask_files: MaskFiles):
-        mpps = []
-        for det, masks in mask_files:
-            mpp = MaskPreprocessor(det)
-            mpp.load_masks(masks)
-            mpps.append(mpp)
-        return mpps
-
-    def gen_plan(
-        self,
-        sample: typing.Union[int, str, dict, list, tuple],
-        plan: typing.Union[int, str, typing.Generator, ScanPlan, list, tuple],
-        robot: bool,
-        poni_file: typing.Optional[PoniFile],
-        mask_files: typing.Optional[MaskFiles]
-    ) -> Plan:
-        """_summary_
-
-        Parameters
-        ----------
-        sample : typing.Union[int, str, dict, list, tuple]
-            _description_
-        plan : typing.Union[int, str, typing.Generator, ScanPlan, list, tuple]
-            _description_
-        robot : bool
-            _description_
-
-        Returns
-        -------
-        Plan
-            _description_
-        """
-        # Translate the index of sample and plan to bluesky plan with metadata
-        plan = xpdacq_composer(
-            self.beamtime,
-            sample,
-            plan,
-            robot=robot,
-            shutter_control=None,
-            dark_strategy=None,
-            auto_load_calib=False
-        )
-        # create one time use cpp if poni_file is given
-        cpps = self._make_cpps(poni_file) if poni_file is not None else self.calib_preprocessors
-        # create one time use mask preprocessor if mask_files are given
-        mpps = self._make_mpps(mask_files) if mask_files is not None else list()
-        for cpp in cpps:
-            plan = cpp(plan)
-        for dpp in self.dark_preprocessors:
-            plan = dpp(plan)
-        for spp in self.shutter_preprocessors:
-            plan = spp(plan)
-        for mpp in mpps:
-            plan = mpp(plan)
-        return plan
-
     def __call__(
         self,
         sample: typing.Union[int, str, dict, list, tuple],
         plan: typing.Union[int, str, typing.Generator, ScanPlan, list, tuple],
-        subs: typing.Union[typing.Callable, dict, list] = None,
+        subs: typing.Union[typing.Callable, list, dict] = None,
         *,
-        poni_file: PoniFile = None,
-        mask_files: MaskFiles = None,
         verify_write: bool = False,
-        dark_strategy: typing.Callable = None,
+        dark_strategy: typing.Callable = periodic_dark,
         robot: bool = False,
         ask_before_run: bool = False,
         **metadata_kw
@@ -291,22 +209,12 @@ class CustomizedRunEngine(RunEngine):
               lists of callables; valid keys are {'all', 'start', 'stop',
               'event', 'descriptor'}
 
-        poni_file: str
-
-            The path to a poni file. This poni file will be read and the data in it will be in the `calib`
-            stream instead of the data in the registered CalibPreprocessors. This is only for one time use.
-
-        mask_files: List[str]
-
-            A list of the paths to mask files. The mask convention is 0 good other bad. The masks will be
-            overlay (sum up) to create a sinlge mask and this mask will be in the `mask` event stream.
-
         verify_write: bool, optional
 
             Double check if the data have been written into database. In general data is written in a lossless
             fashion at the NSLS-II. Therefore, False by default.
 
-        dark_strategy: callable, optional. (deprecated)
+        dark_strategy: callable, optional.
 
             Protocol of taking dark frame during experiment. Default to the logic of matching dark frame and
             light frame with the sample exposure time and frame rate. Details can be found at
@@ -327,30 +235,45 @@ class CustomizedRunEngine(RunEngine):
 
             list of uids (i.e. RunStart Document uids) of run(s)
         """
-        if dark_strategy is not None:
-            raise Warning("dark_strategy is deprecated.")
         if self.md.get("robot", None) is not None:
             raise RuntimeError(
                 "Robot must be specified at call time, not in"
                 "global metadata"
             )
-        # compose the plan
-        final_plan = self.gen_plan(sample, plan, robot, poni_file, mask_files)
-        # normalize the subs
-        _subs = normalize_subs_input(subs) if subs else {}
-        # verify writing files
-        if verify_write:
-            _subs.update({"stop": verify_files_saved})
+        # The CustomizedRunEngine knows about a Beamtime object, and it
+        # interprets integers for 'sample' as indexes into the Beamtime's
+        # lists of Samples from all its Experiments.
+        dark_strategy = dark_strategy if glbl["auto_dark"] else None
+        shutter_control = (
+            xpd_configuration["shutter"], XPD_SHUTTER_CONF["close"]
+        ) if glbl["shutter_control"] else None
+        verbose = 1 if ask_before_run else 0
+        plan = xpdacq_mutator(
+            self._beamtime,
+            sample,
+            plan,
+            robot=robot,
+            shutter_control=shutter_control,
+            dark_strategy=dark_strategy,
+            auto_load_calib=glbl["auto_load_calib"],
+            verbose=verbose
+        )
+        if subs:
+            _subs = normalize_subs_input(subs)
+            # verify writing files
+            if verify_write:
+                _subs.update({"stop": verify_files_saved})
         if robot:
-            metadata_kw.update({'robot': True})
+            metadata_kw.update(robot=True)
         if ask_before_run:
             ip = input("Is this ok? [y]/n")
             if ip.lower() == "n":
                 return
-        return super(CustomizedRunEngine, self).__call__(final_plan, _subs, **metadata_kw)
+        # Execute
+        return super().__call__(plan, subs, **metadata_kw)
 
 
-def xpdacq_composer(
+def xpdacq_mutator(
     beamtime: Beamtime,
     sample: typing.Union[int, dict, typing.List[int]],
     plan: typing.Union[int, Generator, typing.List[int], typing.Generator],
@@ -358,76 +281,92 @@ def xpdacq_composer(
     robot: bool = False,
     shutter_control: typing.Tuple[Device, typing.Any] = None,
     dark_strategy: typing.Callable = None,
-    auto_load_calib: bool = False
+    auto_load_calib: bool = False,
+    verbose: int = 0
 ) -> typing.Generator:
-    """Create a list of plans for an xpd experiment. Used in `~xpdacq.xpdacq.CumstomeizedRunEngine.__call__`.
+    """Create the plan for a xpd experiment. Used in `~xrun.__call__`.
 
     Parameters
     ----------
     beamtime :
+
         The beamtime object.
 
     sample :
+
         If a beamtime object is linked, an integer will be interpreted as the index appears in the ``bt.list()``
         method, corresponding metadata will be passed. A customized dict can also be passed as the sample
         metadata.
 
     plan :
+
         Scan plan. If a beamtime object is linked, an integer will be interpreted as the index appears in the
         ``bt.list()`` method, corresponding scan plan will be A generator or that yields ``Msg`` objects (or an
         iterable that returns such a generator) can also be passed.
 
     robot :
+
         If True, the plan is meant to be using robot.
 
     shutter_control :
+
         A tuple of the shutter device and its close state. The shutter will be closed after the whole scan.
         If None, shutter won't be controlled and no dark will be taken.
 
     dark_strategy :
+
         The strategy how to take the dark frame.
 
     auto_load_calib :
+
         If True, the calibration meta-data will be injected into the run. Else, do nothing.
 
+    verbose :
+
+        0 means no print. 1 means printing the plan out.
+
     Returns
     -------
-    grand_plan :
-        The grand plan to be run by the RunEngine.
+    final_plan :
+
+        A blue sky plan. Used in `~bluesky.RunEngine`.
     """
+    sample, plan = _normalize_sample_plan(sample, plan)
+    # Turn ints into actual samples
+    sample = translate_to_sample(beamtime, sample)
+    # print out
+    if verbose == 1:
+        print_plans(beamtime, sample, plan, robot=robot)
+    # Turn ints into generators
+    plan = translate_to_plan(beamtime, plan, sample)
+    # Collect the plans by contiguous samples and chain them
+    sample, plan = zip(
+        *((s, pchain(*ps)) for s, ps in group_by_sample(sample, plan))
+    )
+    # Make the complete plan by chaining the chained plans
+    total_plan = gen_robot_plans(beamtime, sample, plan) if robot else plan
+    plan = pchain(*total_plan)
     # check wavelength
     warn_wavelength(beamtime)
-    # noramlize the sample and plan to two lists with the same length
-    lst_sample, lst_plan = _normalize_sample_plan(sample, plan)
-    # Turn ints into actual sample dictionary
-    lst_metadata = [translate_to_sample(beamtime, s) for s in lst_sample]
-    # Turn ints into bluesky generators
-    lst_bp_plan = [translate_to_plan(beamtime, p) for p in lst_plan]
-    # Make the complete plan by chaining the chained plans
-    if robot:
-        lst_bp_plan = gen_robot_plans(beamtime, lst_metadata, lst_bp_plan)
-    # Inject the sample metadata
-    lst_bp_plan = [inject_metadata(p, s) for p, s in zip(lst_bp_plan, lst_metadata)]
-    # chain the plans
-    grand_plan = pchain(*lst_bp_plan)
     # shutter control and dark
-    if shutter_control and dark_strategy:
-        grand_plan = dark_strategy(grand_plan)
-        grand_plan = bpp.msg_mutator(grand_plan, _inject_qualified_dark_frame_uid)
-    # Load calibration file
-    if auto_load_calib:
-        grand_plan = bpp.msg_mutator(grand_plan, _inject_calibration_md)
-    # Insert xpdacq md version
-    grand_plan = bpp.msg_mutator(grand_plan, _inject_xpdacq_md_version)
-    # Insert analysis stage tag
-    grand_plan = bpp.msg_mutator(grand_plan, _inject_analysis_stage)
-    # Insert filter metadata
-    grand_plan = bpp.plan_mutator(grand_plan, _inject_filter_positions)
-    # close shutter
     if shutter_control:
         shutter, close_state = shutter_control
-        grand_plan = bpp.finalize_wrapper(grand_plan, close_shutter_at_last(shutter, close_state))
-    return grand_plan
+        # Alter the plan to incorporate dark frames.
+        if dark_strategy:
+            plan = dark_strategy(plan)
+            plan = bpp.msg_mutator(plan, _inject_qualified_dark_frame_uid)
+        # force to close shutter after scan
+        plan = close_shutter_at_last(plan, shutter, close_state)
+    # Load calibration file
+    if auto_load_calib:
+        plan = bpp.msg_mutator(plan, _inject_calibration_md)
+    # Insert xpdacq md version
+    plan = bpp.msg_mutator(plan, _inject_xpdacq_md_version)
+    # Insert analysis stage tag
+    plan = bpp.msg_mutator(plan, _inject_analysis_stage)
+    # Insert filter metadata
+    plan = bpp.msg_mutator(plan, _inject_filter_positions)
+    return plan
 
 
 def close_shutter_at_last(plan: typing.Generator, shutter: Device, close_state: typing.Any) -> typing.Generator:
@@ -462,6 +401,7 @@ def _update_dark_dict_list(name, doc):
         dark_dict_list.append(dark_dict)
         glbl["_dark_dict_list"] = dark_dict_list  # update glbl._dark_dict_list
     else:
+        # FIXME: replace with logging and detailed warning next PR
         print(
             "INFO: dark scan was not successfully executed.\n"
             "gobal dark frame information will not be updated!"
@@ -493,6 +433,7 @@ def take_dark():
     }
     c = bp.count([area_det], md=_md)
     yield from bpp.subs_wrapper(c, {"stop": [_update_dark_dict_list]})
+    # TODO: remove this, since it kinda depends on what happens next?
     print("opening shutter...")
 
 
@@ -586,38 +527,18 @@ def _auto_load_calibration_file(in_scan=True):
         return calib_dict
 
 
-def inject_metadata(plan: typing.Generator, metadata: dict) -> typing.Generator:
-    """Inject the metadata into a plan."""
-
-    def _inject_metadata(msg: Msg):
-        """Inject metadata in the start of the run."""
-        if msg.command == "open_run":
-            msg.kwargs.update(**metadata)
-        return msg
-
-    plan1 = bpp.msg_mutator(plan, _inject_metadata)
-    return plan1
-
-
-def __inject_filter_positions(msg):
-    filter_bank = xpd_configuration["filter_bank"]
-    filter_status = dict()
-    for name in filter_bank.read_attrs:
-        flt = getattr(filter_bank, name)
-        sts = (yield from bps.rd(flt))
-        filter_status[name] = sts
-    print("INFO: Current filter status")
-    for name, status in filter_status.items():
-        print("INFO: {} : {}".format(name, status))
-    msg.kwargs["filter_positions"] = filter_status
-    return (yield msg)
-
-
 def _inject_filter_positions(msg):
     """Inject the filter position in start."""
     if msg.command == "open_run":
-        return __inject_filter_positions(msg), None
-    return None, None
+        filter_bank = xpd_configuration["filter_bank"]
+        filter_status = {
+            name: getattr(filter_bank, name).get() for name in filter_bank.read_attrs
+        }
+        print("INFO: Current filter status")
+        for name, status in filter_status.items():
+            print("INFO: {} : {}".format(name, status))
+        msg.kwargs["filter_positions"] = filter_status
+    return msg
 
 
 def _inject_qualified_dark_frame_uid(msg):
@@ -775,6 +696,7 @@ def set_beamdump_suspender(
 
 def load_sample(position, geometry=None):
     """For robot."""
+    # TODO: I think this can be simpler.
     return (
         yield from single_gen(
             Msg("load_sample", xpd_configuration["robot"], position, geometry)
@@ -784,6 +706,7 @@ def load_sample(position, geometry=None):
 
 def unload_sample():
     """For robot."""
+    # TODO: I think this can be simpler.
     return (
         yield from single_gen(Msg("unload_sample", xpd_configuration["robot"]))
     )
@@ -816,9 +739,9 @@ def robot_wrapper(plan, sample):
 
 def translate_to_sample(
     beamtime: Beamtime,
-    sample: typing.Union[int, str, dict]
+    sample: typing.Union[int, str, dict, list, tuple]
 ) -> typing.Union[dict, typing.List[dict]]:
-    """Translate a sample into a dictionary.
+    """Translate a sample into a list of dict
 
     Parameters
     ----------
@@ -837,43 +760,51 @@ def translate_to_sample(
     sample_md :
         The sample info loaded
     """
-    if isinstance(sample, int):
+    if isinstance(sample, (list, tuple)):
+        sample_md = [translate_to_sample(beamtime, s) for s in sample]
+    elif isinstance(sample, int):
         try:
-            return dict(beamtime.samples.sel(sample))
+            sample_md = list(beamtime.samples.values())[sample]
         except IndexError:
-            raise xpdAcqError(
-                "ERROR: hmm, there is no sample with index `{}`"
+            print(
+                "WARNING: hmm, there is no sample with index `{}`"
                 ", please do `bt.list()` to check if it exists yet".format(
                     sample
                 )
             )
+            sample_md = dict()
     elif isinstance(sample, str):
         try:
-            return dict(beamtime.samples[sample])
+            sample_md = beamtime.samples[sample]
         except KeyError:
-            raise xpdAcqError(
-                "ERROR: hmm, there is no sample with key `{}`"
+            print(
+                "WARNING: hmm, there is no sample with key `{}`"
                 ", please do `bt.list()` to check if it exists yet".format(
                     sample
                 )
             )
-    elif isinstance(sample, OrderedDict):
-        return dict(sample)
-    elif isinstance(sample, dict):
-        return sample
+            sample_md = dict()
     else:
-        raise TypeError(f"The type of sample is {type(sample)}. Expect int, str, dict.")
+        sample_md = sample
+    return sample_md
 
 
-def translate_to_plan(beamtime: Beamtime, plan: typing.Union[int, str, ScanPlan]) -> typing.Generator:
-    """Translate a plan input into a generator.
+def translate_to_plan(beamtime: Beamtime, plan: typing.Union[int, str, ScanPlan, list], sample_md):
+    """Translate a plan input into a generator
 
     Parameters
     ----------
     beamtime : Beamtime
         The BeamTime instance.
 
-    plan : int, str, or dict-like
+    sample_md : list of dict-like
+        Sample metadata. If a beamtime object is linked,
+        an integer will be interpreted as the index appears in the
+        ``bt.list()`` method, corresponding metadata will be passed.
+        A customized dict can also be passed as the sample
+        metadata.
+
+    plan : list, int, str, or dict-like
         Scan plan. If a beamtime object is linked, an integer
         will be interpreted as the index appears in the
         ``bt.list()`` method, corresponding scan plan will be
@@ -883,38 +814,45 @@ def translate_to_plan(beamtime: Beamtime, plan: typing.Union[int, str, ScanPlan]
     Returns
     -------
     plan : generator
-        The generator of messages for the plan。
+        The generator of messages for the plan
+
     """
+    if isinstance(plan, list):
+        plan = [translate_to_plan(beamtime, p, s) for p, s in zip(plan, sample_md)]
     # If a plan is given as a int, look in up in the global registry.
-    if isinstance(plan, int):
-        try:
-            scanplan = beamtime.scanplans.sel(plan)
-            return scanplan.factory()
-        except IndexError:
-            raise xpdAcqError(
-                "ERROR: hmm, there is no scanplan with index `{}`"
-                ", please do `bt.list()` to check if it exists yet".format(
-                    plan
-                )
-            )
-    # If the plan is an xpdAcq 'ScanPlan', make the actual plan.
-    elif isinstance(plan, str):
-        try:
-            scanplan = beamtime.scanplans[plan]
-            return scanplan.factory()
-        except KeyError:
-            raise xpdAcqError(
-                "ERROR: hmm, there is no scanplan with key `{}`"
-                ", please do `bt.list()` to check if it exists yet".format(
-                    plan
-                )
-            )
-    elif isinstance(plan, ScanPlan):
-        return plan.factory()
-    elif isinstance(plan, Generator):
-        return plan
     else:
-        raise TypeError(f"The type of plan is {type(plan)}. Expect int, str, ScanPlan or generator.")
+        if isinstance(plan, int):
+            try:
+                plan = list(beamtime.scanplans.values())[plan]
+            except IndexError:
+                print(
+                    "WARNING: hmm, there is no scanplan with index `{}`"
+                    ", please do `bt.list()` to check if it exists yet".format(
+                        plan
+                    )
+                )
+                return
+        # If the plan is an xpdAcq 'ScanPlan', make the actual plan.
+        elif isinstance(plan, str):
+            try:
+                plan = beamtime.scanplans[plan]
+            except KeyError:
+                print(
+                    "WARNING: hmm, there is no scanplan with key `{}`"
+                    ", please do `bt.list()` to check if it exists yet".format(
+                        plan
+                    )
+                )
+                return
+        elif isinstance(plan, (Generator, ScanPlan)):
+            pass
+        else:
+            raise TypeError(f"The type of plan is {type(plan)}. Expect list, int, str, or dict-like.")
+        if isinstance(plan, ScanPlan):
+            plan = plan.factory()
+        mm = _sample_injector_factory(sample_md)
+        plan = msg_mutator(plan, mm)
+    return plan
 
 
 def _normalize_sample_plan(sample, plan) -> typing.Tuple[list, list]:
@@ -959,7 +897,7 @@ def _normalize_sample_plan(sample, plan) -> typing.Tuple[list, list]:
 def print_plans(
     beamtime: Beamtime,
     sample: typing.List[dict],
-    plan: typing.List[typing.Generator],
+    plan: typing.List[int],
     robot: bool = False
 ) -> None:
     """Print the plan for each sample."""
